import logging, time
import json
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.output_parsers import JsonOutputParser

from .db_core import get_app_settings, get_system_prompt
from .db_agent_ops import save_final_article, add_processed_url, is_url_processed, get_recent_articles
from .agent_helpers import title_similarity, parse_json_output
from .collector_agent import run_collector_agent
from .editor_agent import run_editor_agent
from .llm_setup import initialize_llm
from .db_stats import log_api_call, log_agent_run # Import the logging functions

logger = logging.getLogger(__name__)

DUPLICATE_SIMILARITY_THRESHOLD = 0.85 # Higher threshold for DB duplicate check
SIMILARITY_THRESHOLD = 0.7 # Title similarity threshold for triggering editor (0.0 to 1.0)

# --- Helper function for Manager's specific LLM calls (Evaluation & Keyword Extraction) ---
def ask_manager_llm(llm, token_callback, prompt_text: str, purpose: str) -> dict | None:
    """ Helper to make specific calls to the LLM for Manager tasks like evaluation/keywords """
    agent_name = "Manager"
    logger.debug(f"{agent_name} Agent: Asking LLM for {purpose}. Prompt: {prompt_text[:200]}...")
    try:
        # Directly invoke the LLM with the prompt string.
        # The prompt itself instructs the LLM to return JSON.
        # We will parse the JSON output manually or using a helper.
        # Use a lower temperature for more deterministic JSON output.
        # Pass callback handler to invoke call
        llm_response = llm.invoke(prompt_text, config={"callbacks": [token_callback]})
        success = True # Assume success
        # Get token counts *after* the invoke call
        input_tokens = token_callback.total_input_tokens
        output_tokens = token_callback.total_output_tokens
        # Log API call
        log_api_call(
            api_provider='OpenAI', # Assuming Manager uses OpenAI for these tasks
            service_model=llm.model_name,
            success=success,
            caller_component=f"{agent_name} ({purpose})",
            input_tokens=input_tokens,
            output_tokens=output_tokens,
        )

        raw_output = llm_response.content if hasattr(llm_response, 'content') else str(llm_response)

        # Use the robust JSON parser from agent_helpers
        parsed_json = parse_json_output(raw_output, agent_name=f"{agent_name} ({purpose})")
        return parsed_json if isinstance(parsed_json, dict) else None
    except ValueError as ve: # Config error (e.g., missing key)
        logger.error(f"{agent_name} Agent: Configuration error during LLM call for {purpose}: {ve}")
        log_api_call(api_provider='OpenAI', service_model='N/A', success=False, caller_component=f"{agent_name} ({purpose})", metadata={'error': str(ve)})
    except Exception as e:
        logger.error(f"{agent_name} Agent: Error during LLM call for {purpose}: {e}", exc_info=True)
        return None

# --- Main Manager Agent Function ---
def run_manager_agent(hours: int, tags: list[str], search_depth: int):
    """
    Runs the Manager agent to orchestrate the news pipeline for the given tags.
    It calls the Collector, handles deduplication/merging via the Editor,
    and saves the final articles.
    Args:
        hours: The number of hours to look back for duplicate checks.
        tags: A list of tags defining the scope of the news run.
        search_depth: The number of search results the initial Collector should fetch.
    Includes potential deep dives for noteworthy news if enabled in settings.
    """
    agent_name = "Manager"
    logger.info(f"{agent_name} Agent: Starting run for tags: {tags}")
    if not tags:
        logger.warning(f"{agent_name} Agent: No tags provided. Skipping run.")
        return

    start_time = time.time() # Record start time here
    try:
        # 1. Fetch Settings & Initialize LLM
        app_settings = get_app_settings()
        deep_dive_enabled = app_settings.get('deep_dive_enabled', False)
        deep_dive_depth = app_settings.get('deep_dive_search_depth', 25)
        manager_llm, manager_token_callback = initialize_llm(provider='openai', temperature=0.3) # Get LLM and callback handler
        manager_prompt_template = get_system_prompt(agent_name) # Get base prompt
        if not manager_prompt_template:
             logger.error(f"{agent_name} Agent: Failed to load system prompt. Aborting.")
             return

        # --- Pre-check: Fetch recent articles from DB for duplicate detection ---
        recent_db_articles = get_recent_articles(hours=hours) # Check last N hours
        recent_db_titles = {article['id']: article['title'].lower() for article in recent_db_articles}
        recent_db_summaries = {article['id']: article['summary'] for article in recent_db_articles}
        logger.info(f"{agent_name} Agent: Found {len(recent_db_articles)} articles in DB from the last {hours} hours for duplicate check.")

        # 2. Initial Collection
        collected_articles = run_collector_agent(tags, search_depth)
        if not collected_articles:
            logger.info(f"{agent_name} Agent: Collector found no new articles matching the tags.")
            return

        # --- Filter collected articles against recent DB articles ---
        articles_to_process_further = []
        for collected_article in collected_articles:
            is_potential_duplicate = False
            collected_title_lower = collected_article.get('title', '').lower()
            collected_url = collected_article.get('url')

            for db_article_id, db_title_lower in recent_db_titles.items():
                similarity = title_similarity(collected_title_lower, db_title_lower)
                if similarity >= DUPLICATE_SIMILARITY_THRESHOLD:
                    logger.warning(f"{agent_name} Agent: Potential DB duplicate found (Similarity: {similarity:.2f}) for collected article '{collected_article.get('title')}' (URL: {collected_url}) matching DB article ID {db_article_id} ('{db_title_lower}')")

                    # Confirmation step using LLM
                    compare_prompt = (
                        f"{manager_prompt_template}\n\n"
                        f"--- TASK: Confirm Duplicate Event ---\n"
                        f"Collected Article Content Preview:\n{collected_article.get('content', '')[:1000]}...\n\n"
                        f"Existing Article Summary (DB ID {db_article_id}):\n{recent_db_summaries.get(db_article_id, 'N/A')}\n\n"
                        f"Do these two pieces of text describe the exact same news event or update? Focus on the core event, ignore minor wording differences.\n"
                        f"Respond ONLY with a JSON object like this: {{\"is_duplicate\": true/false}}" # Escaped braces
                    )
                    confirmation_result = ask_manager_llm(manager_llm, manager_token_callback, compare_prompt, f"duplicate confirmation vs DB ID {db_article_id}")

                    if confirmation_result and isinstance(confirmation_result, dict) and confirmation_result.get('is_duplicate') is True:
                        logger.warning(f"{agent_name} Agent: CONFIRMED duplicate. Discarding collected article '{collected_article.get('title')}' (URL: {collected_url}) as it matches DB article {db_article_id}.")
                        if collected_url: add_processed_url(collected_url) # Mark URL as processed
                        is_potential_duplicate = True
                        break # Stop checking against other DB articles once confirmed duplicate
            if not is_potential_duplicate:
                articles_to_process_further.append(collected_article)
        # logger.info(f"{agent_name} Agent: Collector returned {len(collected_articles)} potential articles.") # Removed, less important info

        articles_to_save = [] # Stores tuples: (title, summary, primary_url, source_urls, tags_list)
        processed_indices = set() # Track indices processed (either deep dive or standard)
        processed_urls_in_batch = set() # Track URLs added to DB in this run

        # 3. Process Collected Articles (Evaluate, Deep Dive, or Standard Grouping)
        for i in range(len(articles_to_process_further)): # Use the filtered list
            if i in processed_indices:
                continue # Skip if already part of a group

            article1 = articles_to_process_further[i]
            # Basic validation of collected article structure
            if not all(k in article1 for k in ['url', 'title', 'content']): # Check article1 here
                 logger.warning(f"{agent_name} Agent: Skipping article with missing keys: {article1.get('title', 'N/A')}")
                 processed_indices.add(i)
                 continue

            duplicates_group = [article1] # Start a group with the current article
            processed_indices.add(i)

            article_title = article1.get('title', '')
            article_content_preview = article1.get('content', '')[:1000] # Preview for LLM
            article_url = article1.get('url')

            # --- 3a. Noteworthiness Evaluation (if enabled) ---
            trigger_deep_dive = False
            if deep_dive_enabled:
                eval_prompt = (
                    f"{manager_prompt_template}\n\n"
                    f"--- TASK: Evaluate Noteworthiness ---\n"
                    f"Article Title: {article_title}\n"
                    f"Article Content Preview: {article_content_preview}...\n\n"
                    f"Based *only* on the title and content preview, is this article reporting on potentially exceptional, breaking, or highly significant news for the tags [{', '.join(tags)}]? Is it clearly distinct from routine updates?\n"
                    f"Respond ONLY with a JSON object like this: {{\"noteworthy\": true/false, \"reason\": \"Brief justification (1 sentence)\"}}" # Escaped braces
                )
                evaluation_result = ask_manager_llm(manager_llm, manager_token_callback, eval_prompt, f"noteworthiness evaluation for article {i}")

                if evaluation_result and isinstance(evaluation_result, dict) and evaluation_result.get('noteworthy') is True:
                    logger.info(f"{agent_name} Agent: Article '{article_title}' evaluated as NOTEWORTHY. Reason: {evaluation_result.get('reason', 'N/A')}")
                    trigger_deep_dive = True
                else: # Keep debug log for non-noteworthy
                    logger.debug(f"{agent_name} Agent: Article '{article_title}' not deemed noteworthy for deep dive. Reason: {evaluation_result.get('reason', 'Routine') if isinstance(evaluation_result, dict) else 'Evaluation failed/negative'}")

            # --- 3b. Trigger Deep Dive (if noteworthy) ---
            if trigger_deep_dive:
                # processed_indices.add(i) # Already added when starting loop for article i
                deep_dive_urls = {article_url} # Start with the trigger article's URL

                # Get keywords/query for deep dive
                keyword_prompt = (
                    f"{manager_prompt_template}\n\n"
                    f"--- TASK: Extract Keywords/Query for Deep Dive ---\n"
                    f"Noteworthy Article Title: {article_title}\n"
                    f"Article Content Preview: {article_content_preview}...\n\n"
                    f"Extract the core topic keywords (3-5 specific terms) or a concise search query suitable for finding *more articles strictly about this specific event/topic*.\n"
                    f"Respond ONLY with a JSON object like this: {{\"keywords\": [\"term1\", \"term2\"]}} OR {{\"search_query\": \"concise search phrase\"}}" # Escaped braces
                )
                keyword_result = ask_manager_llm(manager_llm, manager_token_callback, keyword_prompt, f"keyword extraction for article {i}")

                deep_dive_query = None
                if keyword_result and isinstance(keyword_result, dict):
                    if 'keywords' in keyword_result and isinstance(keyword_result['keywords'], list):
                        deep_dive_query = ", ".join(keyword_result['keywords'])
                        # logger.info(f"{agent_name} Agent: Extracted keywords for deep dive: {deep_dive_query}") # Removed, less important info
                    elif 'search_query' in keyword_result and isinstance(keyword_result['search_query'], str):
                        deep_dive_query = keyword_result['search_query']
                        # logger.info(f"{agent_name} Agent: Extracted search query for deep dive: {deep_dive_query}") # Removed, less important info

                if not deep_dive_query:
                    logger.warning(f"{agent_name} Agent: Failed to extract keywords/query for deep dive on '{article_title}'. Falling back to title.")
                    deep_dive_query = article_title # Fallback

                # Run deep dive collector
                logger.info(f"{agent_name} Agent: Starting DEEP DIVE collection (Depth: {deep_dive_depth}) for query: '{deep_dive_query}'")
                deep_dive_articles = run_collector_agent([deep_dive_query], deep_dive_depth) # Pass query as list
                # logger.info(f"{agent_name} Agent: Deep dive collector returned {len(deep_dive_articles)} articles.") # Removed, less important info

                # Prepare for Editor
                articles_for_editor = [article1] + deep_dive_articles # Combine trigger + results
                for dda in deep_dive_articles: deep_dive_urls.add(dda.get('url')) # Collect all URLs

                if len(articles_for_editor) > 0: # Need at least the trigger article
                    logger.info(f"{agent_name} Agent: Sending {len(articles_for_editor)} articles from deep dive to Editor.")
                    edited_result = run_editor_agent(articles_for_editor)
                    if edited_result and edited_result.get('title') and edited_result.get('summary'):
                        # Use the trigger article's URL as primary source for simplicity
                        primary_url = article_url
                        articles_to_save.append((
                            edited_result['title'],
                            edited_result['summary'],
                            primary_url,
                            list(filter(None, deep_dive_urls)), # All source URLs involved
                            tags # Use original agent tags for categorization
                        ))
                        logger.info(f"{agent_name} Agent: Editor returned merged deep dive article: '{edited_result['title']}'")
                    else:
                        logger.warning(f"{agent_name} Agent: Editor failed or returned invalid data for deep dive group. Skipping save.")
                else:
                     logger.warning(f"{agent_name} Agent: No articles to send to editor after deep dive for '{article_title}'.")

            # --- 3c. Standard Processing (if not deep dived) ---
            else:
                # Compare with subsequent articles for similarity
                for j in range(i + 1, len(articles_to_process_further)): # Use the filtered list
                    if j in processed_indices:
                        continue

                    article2 = articles_to_process_further[j]
                    if not all(k in article2 for k in ['url', 'title', 'content']):
                        logger.warning(f"{agent_name} Agent: Skipping article (in comparison) with missing keys: {article2.get('title', 'N/A')}")
                        processed_indices.add(j) # Mark as processed to avoid re-checking
                        continue

                    similarity = title_similarity(article_title, article2.get('title')) # Use article_title here
                    if similarity >= SIMILARITY_THRESHOLD: 
                        # logger.info(f"{agent_name} Agent: Found potential overlap (Similarity: {similarity:.2f}): '{article_title}' AND '{article2.get('title')}'") # Removed, less important info
                        duplicates_group.append(article2)
                        processed_indices.add(j) # Mark as processed since it's in this group

            # Decide whether to call Editor or save directly (for standard processing)
            source_urls = [a.get('url') for a in duplicates_group if a.get('url')]

            if len(duplicates_group) > 1:
                logger.info(f"{agent_name} Agent: Sending {len(duplicates_group)} overlapping articles to Editor.")
                edited_result = run_editor_agent(duplicates_group)
                if edited_result and edited_result.get('title') and edited_result.get('summary'):
                    # Use the first URL as the primary source for simplicity, or None
                    primary_url = source_urls[0] if source_urls else None
                    articles_to_save.append((
                        edited_result['title'],
                        edited_result['summary'],
                        primary_url,
                        source_urls, # Keep track of all original URLs for processing later
                        tags # Use the original query tags for the merged article
                    ))
                    logger.info(f"{agent_name} Agent: Editor returned merged article: '{edited_result['title']}'")
                else:
                    logger.warning(f"{agent_name} Agent: Editor failed or returned invalid data for a group of {len(duplicates_group)} articles. Skipping group.")
            else:
                # No duplicates found for article1, save it directly
                # Use scraped content as summary, maybe truncate or ask LLM for brief summary later?
                # For now, truncate content. A better approach might be a quick summarization call.
                summary = article1.get('content', 'No Content Provided') # Use article1 content
                summary = (summary[:500] + '...') if len(summary) > 500 else summary # Truncate long content
                # logger.info(f"{agent_name} Agent: Preparing single standard article for saving: '{article_title}'") # Removed, less important info
                articles_to_save.append((
                    article_title or 'No Title Provided',
                    summary,
                    article_url,
                    list(filter(None, source_urls)), # Should just be article_url
                    tags # Use original query tags
                ))

        # 4. Save Final Articles and Mark URLs as Processed
        saved_count = 0
        # logger.info(f"{agent_name} Agent: Processing {len(articles_to_save)} final articles/groups for saving.") # Removed, less important info
        
        # --- Deduplication based on primary_url before saving ---
        final_articles_by_url = {}
        for article_data in articles_to_save:
            _, _, primary_url, _, _ = article_data # Unpack to get the primary URL
            if primary_url and primary_url not in final_articles_by_url:
                final_articles_by_url[primary_url] = article_data
            elif not primary_url:
                 logger.warning(f"{agent_name} Agent: Skipping article with missing primary_url during final save step: {article_data[0]}")

        # logger.info(f"{agent_name} Agent: Saving {len(final_articles_by_url)} articles after URL deduplication.") # Removed, less important info
        for title, summary, primary_url, original_urls, article_tags in final_articles_by_url.values():
            # Double-check URLs haven't been processed *since the collector ran*
            # This is a safety check against race conditions if runs were very frequent.
            valid_original_urls = [url for url in original_urls if url] # Filter out None/empty URLs
            urls_to_mark_processed = []
            article_already_processed = True # Assume processed unless a URL is new

            for url in valid_original_urls:
                if not is_url_processed(url):
                    urls_to_mark_processed.append(url)
                    article_already_processed = False

            # Use the first *new* URL as the primary source_url in final_articles table
            # Or fall back to the first original URL if somehow all were processed (shouldn't happen due to check above)
            final_source_url = urls_to_mark_processed[0] if urls_to_mark_processed else primary_url

            # Ensure title and summary are not empty before saving
            if not title or not summary:
                logger.warning(f"{agent_name} Agent: Skipping article with empty title or summary. Primary URL: {final_source_url}")
                continue

            success = save_final_article(title, summary, final_source_url, article_tags)
            if success:
                saved_count += 1
                # Mark all newly processed original URLs associated with this saved article as processed
                for url in urls_to_mark_processed:
                    if add_processed_url(url): # Add to DB
                       processed_urls_in_batch.add(url)
                    else:
                       logger.warning(f"{agent_name} Agent: Failed to mark URL as processed in DB: {url}")
            else:
                logger.error(f"{agent_name} Agent: Failed to save final article to DB: '{title}'")

        logger.info(f"{agent_name} Agent: Finished run for tags {tags}. Saved {saved_count} final articles. Marked {len(processed_urls_in_batch)} unique URLs as processed in this run.")

    except Exception as e:
        logger.error(f"{agent_name} Agent: Unhandled error during execution for tags {tags}: {e}", exc_info=True)
    finally:
        end_time = time.time()
        log_agent_run(f"{agent_name}:{','.join(tags)}", start_time, end_time, True) # Log the run with tags
